package provide BSMainWindow 1.0

package require BSLoginWindow 1.0
package require BSFiles 1.0
package require BSInventory 1.0
package require Brickshelf 1.0
package require ZipFiles 1.0

namespace eval BSMainWindow {

# external procedures
variable procs
array set procs {
	username {}
	logout {}
}

# upload configuration
variable config
array set config {
	inparent ""
	inexisting ""
}

# key widgets
variable widgets
array set widgets {
	
	nameentry {}
	newmenu {}
	newfmb {}
	descentry {}
	tagentry {}
	tagbutton {}
	tagw {}
	
	oldmenu {}
	oldfmb {}
	
	logoutb {}
	uploadb {}
}

variable status {}
variable busy 0

# public procedures
namespace export Setup MakeMainWindow IsBusy

#
# Setup
#
# Parameters:
#	usernameProc, procedure that returns authenticated username
#	logoutProc, procedure to call to log out (return to login)
#
# Return:
#	none
#
proc Setup { usernameProc logoutProc } {
	variable procs
	array set procs [list \
		username $usernameProc \
		logout $logoutProc]
	::BSMainWindow::Reset
}

#
# Reset
#
# Apply default main window options, excepting inventory menus.
#
# Parameters:
#	none
#
# Return:
#	none
#
proc Reset {} {
	variable config
	array set config {
		newfolder 1
		
		name ""
		description ""
		tags ""
	}
}

#
# MakeMainWindow
#
# Parameters:
#	username, logged in username (for display purposes)
#
# Return:
#	none
#
proc MakeMainWindow {} {
	variable procs
	
	# clear window
	foreach child [winfo children .] {
		destroy $child
	}
	
	# window properties
	wm title . [format "Brickshelver: %s" [$procs(username)]]
	wm resizable . 1 1
	wm geometry . 480x300
	wm minsize . 480 300
	
	# window panes
	panedwindow .p -orient horizontal -showhandle 1
	pack .p -expand true -fill both
	
	# populate panes
	::BSMainWindow::MakeFileListFrame [frame .p.f]
	::BSMainWindow::MakeConfigurationFrame [frame .p.c]
	
	# install panes
	.p add .p.f -minsize 150
	.p add .p.c	-minsize 300
}

#
# MakeFileListFrame
#
# Parameters:
#	f, container widget
#
# Return:
#	none
#
proc MakeFileListFrame { f } {
	
	# file list -selectmode browse
	set l [::BSMainWindow::ScrolledListbox $f.flist -listvariable [::BSFiles::Reference] -selectmode extended -bg white -relief sunken]
	pack $f.flist -expand yes -fill both
	
	# file list controls
	frame $f.ctrl
	button $f.ctrl.add -text "+" -command ::BSMainWindow::SelectFiles
	button $f.ctrl.rm -text "-" -command [list ::BSMainWindow::DeleteFiles $l]
	pack $f.ctrl.add $f.ctrl.rm -side left
	pack $f.ctrl -fill x -anchor s
}

#
# ScrolledListbox
#
# Parameters:
#	f, path for new listbox group
#	args, additional listbox options
#
# Return:
#	path of actual listbox widget
#
proc ScrolledListbox { f args } {
	frame $f
	listbox $f.list \
			-xscrollcommand [list ::BSMainWindow::ScrollSet $f.xscroll [list grid $f.xscroll -row 1 -column 0 -sticky we]] \
			-yscrollcommand [list ::BSMainWindow::ScrollSet $f.yscroll [list grid $f.yscroll -row 0 -column 1 -sticky ns]]
	eval {$f.list configure} $args
	scrollbar $f.xscroll -orient horizontal -command [list $f.list xview]
	scrollbar $f.yscroll -orient vertical -command [list $f.list yview]
	grid $f.list -sticky news
	grid rowconfigure $f 0 -weight 1
	grid columnconfigure $f 0 -weight 1
	return $f.list
}

#
# ScrollSet
#
# There is an "autoscroll" package in Tklib.
#
# Parameters:
#	scrollbar,
#	geoCmd,
#	offset,
#	size,
#
# Return:
#	none
#
proc ScrollSet {scrollbar geoCmd offset size} {
	if {$offset != 0.0 || $size != 1.0} {
		eval $geoCmd
	}
	$scrollbar set $offset $size
}
	
#
# SelectFiles
#
# Pick files to add to the file list.
#
# Parameters:
#	none
#
# Return:
#	none
#
proc SelectFiles {} {
	# for file type restriction, obtain filetypes list from BSFiles
	set result [tk_getOpenFile -title "Upload Files" -multiple true]
	if {$result == {}} {
		return
	}
	::BSFiles::AddFiles $result
}

#
# DeleteFiles
#
# Remove selected files from the file list.
#
# Parameters:
#	l, list widget to get selection from
#
# Return:
#	none
#
proc DeleteFiles { l } {
	set indices [$l curselection]
	if {$indices == {}} {
		return
	}
	::BSFiles::RemoveFiles $indices
}

#
# MakeConfigurationFrame
#
# Parameters:
#	f, container widget
#
# Return:
#	none
#
proc MakeConfigurationFrame { f } {
		
	# destination folder setup
	labelframe .p.c.folders -text "Folder"
	::BSMainWindow::MakeFolderConfigFrame .p.c.folders
	pack .p.c.folders -fill x
		
	# main controls
	frame .p.c.control
	::BSMainWindow::MakeConfigControlFrame .p.c.control
	pack .p.c.control -fill x
}

#
# MakeFolderConfigFrame
#
# Parameters:
#	f, container widget
#
# Return:
#	none
#
proc MakeFolderConfigFrame { f } {		
	
	# new folder
	::BSMainWindow::MakeNewFolderFrame [frame $f.new]
	pack $f.new -fill x
	
	# existing folders, if any
	if {[::BSInventory::Count] > 0} {
		::BSMainWindow::MakeExistingFolderFrame [frame $f.old]
		pack $f.old -fill x
	}
}

#
# MakeNewFolderFrame
#
# Parameters:
#	f, container widget
#
# Return:
#	none
#
proc MakeNewFolderFrame { f } {
	variable config
	variable widgets
	
	# if there are no folders yet, creating a new one is the only option
	if {[::BSInventory::Count] > 0} {
		radiobutton $f.new -text "New folder:" -value 1 -variable ::BSMainWindow::config(newfolder)
	} else {
		label $f.new -text "New folder:"
	}
	set widgets(nameentry) [entry $f.name -bg white -relief sunken -textvariable ::BSMainWindow::config(name)]
	
	# create new folder within which folder, including account root
	label $f.parentl -text "In:"
	set config(inparent) "/"
	set widgets(newmenu) [eval {tk_optionMenu $f.parent ::BSMainWindow::config(inparent) $config(inparent)} [::BSInventory::Paths]]
	set widgets(newfmb) $f.parent
	$f.parent config -anchor w -takefocus 1

	# descriptive text
	label $f.descl -text "Description:"
	set widgets(descentry) [entry $f.desc -bg white -relief sunken -textvariable ::BSMainWindow::config(description)]

	# keywords, labels, tags, whatever
	label $f.tagsl -text "Tags:"
	set tf [frame $f.tagctrl]
	set widgets(tagentry) [entry $tf.tags -bg white -relief sunken -textvariable ::BSMainWindow::config(tags) -validatecommand {::BSMainWindow::UpdateTags %P} -validate key]
	set widgets(tagbutton) [button $tf.stdtags -text "+" -command ::BSMainWindow::MakeTagWindow]
	
	# grid the tag field+stdtag button combo
	grid $tf.tags $tf.stdtags
	grid $tf.tags -sticky we
	grid columnconfigure $tf 0 -weight 1
	
	# grid the new folder fields
	grid $f.new $f.name
	grid $f.name -sticky we
	grid $f.parentl $f.parent
	grid $f.parentl -sticky e
	grid $f.parent -sticky we
	grid $f.descl $f.desc
	grid $f.descl -sticky e
	grid $f.desc -sticky we
	grid $f.tagsl $f.tagctrl
	grid $f.tagsl -sticky e
	grid $f.tagctrl -sticky we
	grid columnconfigure $f 1 -weight 1
}

#
# RebuildNewFolderMenu
#
# Parameters:
#	none
#
# Return:
#	none
#
proc RebuildNewFolderMenu {} {
	variable widgets
	variable config
	
	# identify original menu button - and zap it
	set mb [winfo parent $widgets(newmenu)]
	set gw [grid info $mb]
	destroy $mb
	
	# build a perhaps-updated optionMenu in its place
	set ::BSMainWindow::config(inparent) "/"
	set widgets(newmenu) [eval {tk_optionMenu $mb ::BSMainWindow::config(inparent) $::BSMainWindow::config(inparent)} [::BSInventory::Paths]]
	$mb config -anchor w -takefocus 1
	
	# use the original menu's grid settings
	eval {grid $mb} $gw
}

#
# MakeExistingFolderFrame
#
# Parameters:
#	f, container widget
#
# Return:
#	none
#
proc MakeExistingFolderFrame { f } {
	variable config
	variable widgets
	
	radiobutton $f.old -text "Existing folder:" -value 0 -variable ::BSMainWindow::config(newfolder)
	
	# menu of existing folders
	set dirs [::BSInventory::Paths]
	set config(inexisting) [lindex $dirs 0]
	set widgets(oldmenu) [eval {tk_optionMenu $f.parent ::BSMainWindow::config(inexisting) $config(inexisting)} [lrange $dirs 1 end]]
	set widgets(oldfmb) $f.parent
	$f.parent config -anchor w -takefocus 1
	
	# grid existing folder fields
	grid $f.old $f.parent
	grid $f.parent -sticky we
	grid columnconfigure $f 1 -weight 1
}

#
# RebuildExistingFolderMenu
#
# Parameters:
#	none
#
# Return:
#	none
#
proc RebuildExistingFolderMenu {} {
	variable widgets
	variable config
	
	# identify original menu button - and zap it
	set mb [winfo parent $widgets(oldmenu)]
	set gw [grid info $mb]
	destroy $mb
	
	# build a perhaps-updated optionMenu in its place
	set dirs [::BSInventory::Paths]
	set ::BSMainWindow::config(inexisting) [lindex $dirs 0]
	set widgets(oldmenu) [eval {tk_optionMenu $mb ::BSMainWindow::config(inexisting) $::BSMainWindow::config(inexisting)} [lrange $dirs 1 end]]
	$mb config -anchor w -takefocus 1

	# use the original menu's grid settings
	eval {grid $mb} $gw
}

#
# MakeConfigControlFrame
#
# Parameters:
#	f, container widget
#
# Return:
#	none
#
proc MakeConfigControlFrame { f } { 
	variable status
	variable widgets
	
	# status field - an important widget
	entry $f.statusl -state readonly -relief flat -textvariable ::BSMainWindow::status -takefocus 0
	
	# main options
	set widgets(logoutb) [button $f.logout -text "Logout" -command ::BSMainWindow::Logout]
	set widgets(uploadb) [button $f.upload -text "Upload" -command ::BSMainWindow::Upload]
	
	pack $f.statusl -side left -fill x -expand true
	pack $f.upload -side right
	pack  $f.logout -side right
}

#
# DisableMainControls
#
# Parameters:
#	none
#
# Return:
#	none
#
proc DisableMainControls {} {
	variable widgets
	$widgets(logoutb) config -state disabled
	$widgets(uploadb) config -state disabled
}

#
# EnableMainControls
#
# Parameters:
#	none
#
# Return:
#	none
#
proc EnableMainControls {} {
	variable widgets
	$widgets(logoutb) config -state normal
	$widgets(uploadb) config -state normal
}

#
# Logout
#
# Perform "clean-up" tasks that need to be done before
# ceding control to the external "logout" procedure.
#
# Parameters:
#	none
#
# Return:
#	none
#
proc Logout {} {
	variable widgets
	variable procs
	::BSMainWindow::Status
	if {$widgets(tagw) != {}} {
		::BSMainWindow::DeleteTagWindow
	}
	$procs(logout)
}

#
# Status
#
# Parameters:
#	txt, optional status message to display
#
# Return:
#	none
#
proc Status { {txt {}} } {
	variable status
	set status $txt
	update idletasks
}

#
# Upload
#
# There are some aesthetic issues I need to address during uploading,
# such as: input fields should probably be disabled; errors should be
# reported in more detail (local or network); perhaps a visual meter?
#
# Parameters:
#	none
#
# Return:
#	none
#
proc Upload {} {
	variable config
	variable busy
	
	# have to have something to upload
	if {[::BSFiles::Count] <= 0} {
		return
	}
	
	set busy 1
	::BSMainWindow::DisableMainControls
	
	# create new folder, if necessary
	if {$config(newfolder)} {
		if {![ValidateName $config(name)]} {
			# focus and select name
			Status "Invalid folder name"
			set busy 0
			::BSMainWindow::EnableMainControls
			return
		}
		if {![ValidateInfo $config(description)]} {
			Status "Invalid description"
			set busy 0
			::BSMainWindow::EnableMainControls
			return
		}
		if {![ValidateInfo $config(tags)]} {
			Status "Invalid tags"
			set busy 0
			::BSMainWindow::EnableMainControls
			return
		}
		set inparent [::BSInventory::ID $config(inparent)]
		if {[catch {::Brickshelf::CreateFolder \
				$config(name) \
				$config(description) \
				$config(tags) \
				::BSMainWindow::Status \
				$inparent} parent]} {
			set busy 0
			::BSMainWindow::EnableMainControls
			ErrorMsg $parent 
		}
		
		# save folder & update menus. p is a hack.
		set p ""
		if {$config(inparent) ne "/"} {
			set p $config(inparent)
		}
		::BSInventory::Add [format "%s/%s" $p $config(name)] $parent
		::BSInventory::Save
		RebuildExistingFolderMenu
		RebuildNewFolderMenu
		
	} else {
		set parent [::BSInventory::ID $config(inexisting)]
	}
		
	# zip the files
	set zipdata ""
	if {[catch {::ZipFiles::ZipFiles [::BSFiles::Files] zipdata ::BSMainWindow::Status} bytelength]} {
		set busy 0
		::BSMainWindow::EnableMainControls
		ErrorMsg $bytelength
	}
	
	# upload the file
	if {[catch {::Brickshelf::UploadFile "foo.zip" $parent ::BSMainWindow::Status zipdata} err]} {
		set busy 0
		::BSMainWindow::EnableMainControls
		ErrorMsg $err
	}
	
	# fresh start
	::BSMainWindow::Reset
	::BSFiles::RemoveFiles
	
	set busy 0
	::BSMainWindow::EnableMainControls
}

#
# IsBusy
#
# Parameters:
#	none
#
# Return:
#	1 upload in progress
#	0 inactive
#

proc IsBusy {} {
	variable busy
	return $busy
}

#
# ErrorMsg
#
# Parameters:
#	msg, error message to display
#
# Return:
#	none (returns from caller)
#
proc ErrorMsg { msg } {
	tk_messageBox -message $msg -type ok
	return -code return
}

#
# MakeTagWindow
#
# Parameters:
#	none
#
# Return:
#	none
#
proc MakeTagWindow {} {
	variable config
	variable widgets
	
	# toggle if already open
	if {$widgets(tagw) != {}} {
		DeleteTagWindow
		return
	}
	
	$widgets(tagbutton) config -text "-"
	
	# list of current tags
	set taglist [split [string tolower $config(tags)]]
	
	# create tag window
	set widgets(tagw) [toplevel .t]
	wm title $widgets(tagw) "Tags"
	wm resizable . 0 0
	wm protocol $widgets(tagw) WM_DELETE_WINDOW ::BSMainWindow::DeleteTagWindow
	
	# create tag options
	set row 0
	set col 0
	foreach tag {MOC Bionicle Train Castle Technic Space 
			Town Mecha Club CAD Multimedia Instructions 
			LEGOLAND Auction People Avatar} {
		
		# create checkbox for this widget
		set w [format "%s.%s" $widgets(tagw) [string tolower $tag]]
		checkbutton $w -text $tag -command [list ::BSMainWindow::ToggleTag $tag]
		
		# check this tag if it is present in field
		if {[lsearch $taglist [string tolower $tag]] != -1} {
			$w select
		}
		
		# place checkbox in grid
		grid $w -row $row -column $col -sticky w	
		if {[incr col] > 1} {
			set col 0
			incr row
		}
	}
}

#
# DeleteTagWindow
#
# Parameters:
#	none
#
# Return:
#	none
#
proc DeleteTagWindow {} {
	variable widgets
	if {$widgets(tagw) == {}} {
		return
	}
	$widgets(tagbutton) config -text "+"
	destroy $widgets(tagw)
	set widgets(tagw) {}
}

#
# ToggleTag
#
# Doesn't handle duplicate tags!
#
# Parameters:
#	tag, term to add or remove to the list of tags
#
# Return:
#	none
#
proc ToggleTag { tag } {
	variable config

	set ltags [split [string tolower $config(tags)]]
	
	# is this tag already in the list?
	set tagi [lsearch $ltags [string tolower $tag]]
		
	if {$tagi == -1} {
		# add new tag
		set config(tags) [concat $config(tags) $tag]
	} else {
		# remove existing tag
		set config(tags) [join [lreplace [split $config(tags)] $tagi $tagi]]
	}
}

#
# UpdateTags
#
# Parameters:
#	txt, string possibly containing standard tags
#
# Return:
#	1 txt approved
#		(triggered as a field validator)
#
proc UpdateTags { txt } {
	variable widgets
	if {$widgets(tagw) == {}} {
		return 1
	}
	set taglist [split [string tolower $txt]]
	foreach tag {MOC Bionicle Train Castle Technic Space 
			Town Mecha Club CAD Multimedia Instructions 
			LEGOLAND Auction People Avatar} {
		set w [format "%s.%s" $widgets(tagw) [string tolower $tag]]
		if {[lsearch $taglist [string tolower $tag]] != -1} {
			$w select
		} else {
			$w deselect
		}
	}
	return 1
}

#
# ValidateName
#
# Parameters:
#	name, candidate folder name
#
# Return:
#	1 meets requirements
#	0 fails requirements
#
proc ValidateName { name } {
	# Intended to test at usage time, not during edit
	# Names may only contain A-Z,a-z,0-9 and '-'  (character sets)
	# Names may not start or end with a '-'       (- only in middle)
	# Names must be at least 2 characters long    (^ and $)
	# is there a maximum length?
	if {[regexp {^[a-zA-Z0-9][a-zA-Z0-9\-]*[a-zA-Z0-9]$} $name]} {
		return 1
	} else {
		return 0
	}
}

#
# ValidateInfo
#
# Parameters:
#	info, candidate description/tag string
#
# Return:
#	1 meets requirements
#	0 fails requirements
#
proc ValidateInfo { txt } {
	# [A-Z][a-z][0-9][-][_][@][.][ ]
	# plus comma, and maybe others?
	# We require minimum length 1 (perhaps that should be restricted to non-whitespace?)
	# is there a maximum length?
	if {[regexp {^[a-zA-Z0-9\-_.@ ,]+$} $txt]} {
		return 1
	} else {
		return 0
	}
}

}
